from argparse import ArgumentParser, HelpFormatter
from logging import getLogger, StreamHandler, Formatter
from os import environ, getcwd
from pathlib import Path
from re import compile

from snowddl.fernet.wrapper import FernetWrapper


def init_logger():
    logger = getLogger("snowddl")
    logger.setLevel("INFO")

    formatter = Formatter("%(asctime)s - %(levelname)s - %(message)s")
    formatter.default_msec_format = "%s.%03d"

    handler = StreamHandler()
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger


def entry_point():
    parser = build_parser()
    args = vars(parser.parse_args())
    wrapper = FernetWrapper(args.get("k"))

    if args["action"] == "generate-key":
        action_generate_key(args, wrapper)
    elif args["action"] == "encrypt":
        action_encrypt(args, wrapper)
    elif args["action"] == "decrypt":
        action_decrypt(args, wrapper)
    elif args["action"] == "rotate":
        action_rotate(args, wrapper)
    elif args["action"] == "config-encrypt":
        action_config_encrypt(args, wrapper)
    elif args["action"] == "config-decrypt":
        action_config_decrypt(args, wrapper)
    elif args["action"] == "config-rotate":
        action_config_rotate(args, wrapper)


def build_parser():
    formatter = lambda prog: HelpFormatter(prog, max_help_position=32)

    parser = ArgumentParser(
        prog="snowddl-fernet",
        description="Utility functions to encrypt secrets for SnowDDL config with Fernet",
        formatter_class=formatter,
    )

    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True

    # Action: generate-key
    generate_key_subparser = subparsers.add_parser("generate-key", help="Generate a fresh encryption key")

    generate_key_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    generate_key_subparser.add_argument(
        "--prepend",
        help="Prepend newly generated key to existing keys",
        default=False,
        action="store_true",
    )

    generate_key_subparser.add_argument(
        "--export",
        help="Output key(s) as export env command for CLI",
        default=False,
        action="store_true",
    )

    # Action: encrypt
    encrypt_subparser = subparsers.add_parser("encrypt", help="Encrypt value with first key")

    encrypt_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    encrypt_subparser.add_argument(
        "value",
        help="String value",
        metavar="VALUE",
    )

    # Action: decrypt
    decrypt_subparser = subparsers.add_parser("decrypt", help="Decrypt value with any key")

    decrypt_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    decrypt_subparser.add_argument(
        "value",
        help="Encrypted value",
        metavar="VALUE",
    )

    # Action: rotate
    rotate_subparser = subparsers.add_parser("rotate", help="Rotate value encrypted with any key")

    rotate_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    rotate_subparser.add_argument(
        "value",
        help="Encrypted value to rotate",
        metavar="VALUE",
    )

    # Action: config-encrypt
    config_encrypt_subparser = subparsers.add_parser("config-encrypt", help="Encrypy all values in config")

    config_encrypt_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    config_encrypt_subparser.add_argument(
        "-c",
        help="Path to config directory (default: current directory)",
        metavar="CONFIG_PATH",
        default=getcwd(),
    )

    # Action: config-decrypt
    config_decrypt_subparser = subparsers.add_parser("config-decrypt", help="Decrypt all values in config")

    config_decrypt_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    config_decrypt_subparser.add_argument(
        "-c",
        help="Path to config directory (default: current directory)",
        metavar="CONFIG_PATH",
        default=getcwd(),
    )

    # Action: config-rotate
    config_rotate_subparser = subparsers.add_parser("config-rotate", help="Rotate all values in config")

    config_rotate_subparser.add_argument(
        "-k",
        help="All current encryption keys separated by comma (<key1>,<key2>,<key3>)",
        metavar="ENCRYPTION_KEYS",
        default=environ.get(FernetWrapper.ENV_ENCRYPTION_KEYS),
    )

    config_rotate_subparser.add_argument(
        "-c",
        help="Path to config directory (default: current directory)",
        metavar="CONFIG_PATH",
        default=getcwd(),
    )

    return parser


def action_generate_key(args, wrapper: FernetWrapper):
    all_keys = [wrapper.generate_key()]

    if args.get("prepend"):
        all_keys.extend(wrapper.key_sequence)

    if args.get("export"):
        print(f"export {wrapper.ENV_ENCRYPTION_KEYS}={','.join(all_keys)}")
    else:
        print(",".join(all_keys))


def action_encrypt(args, wrapper: FernetWrapper):
    print(wrapper.encrypt(args["value"]))


def action_decrypt(args, wrapper: FernetWrapper):
    print(wrapper.decrypt(args["value"]))


def action_rotate(args, wrapper: FernetWrapper):
    print(wrapper.rotate(args["value"]))


def action_config_encrypt(args, wrapper: FernetWrapper):
    logger = init_logger()
    regexp = compile(r"!encrypt\s+\"?(.+?)\"?\n")

    for file in get_config_files_generator(args):
        original_text = file.read_text(encoding="utf-8")
        updated_text, number_of_sub = regexp.subn(lambda m: f"!decrypt {wrapper.encrypt(m[1])}\n", original_text)

        if number_of_sub > 0:
            file.write_text(updated_text, encoding="utf-8")
            logger.info(f"Updated file [{file}] with [{number_of_sub}] substitutions")


def action_config_decrypt(args, wrapper: FernetWrapper):
    logger = init_logger()
    regexp = compile(r"!decrypt\s+\"?(.+?)\"?\n")

    for file in get_config_files_generator(args):
        original_text = file.read_text(encoding="utf-8")
        updated_text, number_of_sub = regexp.subn(lambda m: f'!encrypt "{wrapper.decrypt(m[1])}"\n', original_text)

        if number_of_sub > 0:
            file.write_text(updated_text, encoding="utf-8")
            logger.info(f"Updated file [{file}] with [{number_of_sub}] substitutions")


def action_config_rotate(args, wrapper: FernetWrapper):
    logger = init_logger()
    regexp = compile(r"!decrypt\s+\"?(.+?)\"?\n")

    for file in get_config_files_generator(args):
        original_text = file.read_text(encoding="utf-8")
        updated_text, number_of_sub = regexp.subn(lambda m: f"!decrypt {wrapper.rotate(m[1])}\n", original_text)

        if number_of_sub > 0:
            file.write_text(updated_text, encoding="utf-8")
            logger.info(f"Updated file [{file}] with [{number_of_sub}] substitutions")


def get_config_files_generator(args):
    path = Path(args["c"])

    if not path.is_dir():
        raise ValueError(f"Config path [{path}] is not a directory")

    for file in path.glob("**/*.yaml"):
        yield file



from pathlib import Path

from .ident import SchemaObjectIdent
from ..model import BaseModelWithConfig


class StageWithPath(BaseModelWithConfig):
    stage_name: SchemaObjectIdent
    path: str


class StageUploadFile(BaseModelWithConfig):
    local_path: Path
    stage_path: str



from ._parsed_file import ParsedFile
from ._scanner import DirectoryScanner
from .abc_parser import AbstractParser
from .account_params import AccountParameterParser
from .account_policy import AccountPolicyParser
from .aggregation_policy import AggregationPolicyParser
from .alert import AlertParser
from .authentication_policy import AuthenticationPolicyParser
from .backup_policy import BackupPolicyParser
from .backup_set import BackupSetParser
from .business_role import BusinessRoleParser
from .database import DatabaseParser
from .database_role import DatabaseRoleParser
from .dynamic_table import DynamicTableParser
from .event_table import EventTableParser
from .external_access_integration import ExternalAccessIntegrationParser
from .external_function import ExternalFunctionParser
from .external_table import ExternalTableParser
from .file_format import FileFormatParser
from .function import FunctionParser
from .hybrid_table import HybridTableParser
from .iceberg_table import IcebergTableParser
from .materialized_view import MaterializedViewParser
from .masking_policy import MaskingPolicyParser
from .network_policy import NetworkPolicyParser
from .network_rule import NetworkRuleParser
from .outbound_share import OutboundShareParser
from .permission_model import PermissionModelParser
from .pipe import PipeParser
from .placeholder import PlaceholderParser
from .procedure import ProcedureParser
from .projection_policy import ProjectionPolicyParser
from .resource_monitor import ResourceMonitorParser
from .row_access_policy import RowAccessPolicyParser
from .schema import SchemaParser
from .secret import SecretParser
from .semantic_view import SemanticViewParser
from .sequence import SequenceParser
from .stage import StageParser
from .stream import StreamParser
from .table import TableParser
from .task import TaskParser
from .technical_role import TechnicalRoleParser
from .user import UserParser
from .view import ViewParser
from .warehouse import WarehouseParser


default_parse_sequence = [
    AccountParameterParser,
    # --
    AggregationPolicyParser,
    AuthenticationPolicyParser,
    MaskingPolicyParser,
    NetworkPolicyParser,
    ProjectionPolicyParser,
    RowAccessPolicyParser,
    ResourceMonitorParser,
    AccountPolicyParser,
    BackupPolicyParser,
    BackupSetParser,
    # --
    WarehouseParser,
    DatabaseParser,
    SchemaParser,
    SecretParser,
    NetworkRuleParser,
    ExternalAccessIntegrationParser,
    FileFormatParser,
    StageParser,
    SequenceParser,
    FunctionParser,
    ExternalFunctionParser,
    ProcedureParser,
    TableParser,
    EventTableParser,
    HybridTableParser,
    IcebergTableParser,
    DynamicTableParser,
    ExternalTableParser,
    MaterializedViewParser,
    ViewParser,
    SemanticViewParser,
    PipeParser,
    StreamParser,
    TaskParser,
    AlertParser,
    # --
    DatabaseRoleParser,
    OutboundShareParser,
    TechnicalRoleParser,
    BusinessRoleParser,
    UserParser,
]


singledb_parse_sequence = [
    AggregationPolicyParser,
    MaskingPolicyParser,
    ProjectionPolicyParser,
    RowAccessPolicyParser,
    BackupPolicyParser,
    BackupSetParser,
    # --
    DatabaseParser,
    SchemaParser,
    SecretParser,
    NetworkRuleParser,
    FileFormatParser,
    StageParser,
    SequenceParser,
    FunctionParser,
    ExternalFunctionParser,
    ProcedureParser,
    TableParser,
    EventTableParser,
    HybridTableParser,
    IcebergTableParser,
    DynamicTableParser,
    ExternalTableParser,
    MaterializedViewParser,
    ViewParser,
    SemanticViewParser,
    PipeParser,
    StreamParser,
    TaskParser,
    AlertParser,
    # --
    DatabaseRoleParser,
]



from snowddl.blueprint import IcebergTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class IcebergTableResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.ICEBERG_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW ICEBERG TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Currently only external iceberg tables are supported
            if r["iceberg_table_type"] != "UNMANAGED":
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(IcebergTableBlueprint)

    def create_object(self, bp: IcebergTableBlueprint):
        create_query = self.engine.query_builder()
        common_query = self._build_common_unmanaged_iceberg_table_sql(bp)

        create_query.append(
            "CREATE ICEBERG TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append_nl(common_query)

        self.engine.execute_safe_ddl(create_query)
        self.engine.execute_safe_ddl(
            "ALTER ICEBERG TABLE {full_name:i} SET COMMENT = {comment}",
            {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: IcebergTableBlueprint, row: dict):
        replace_query = self.engine.query_builder()
        common_query = self._build_common_unmanaged_iceberg_table_sql(bp)

        if not common_query.compare_short_hash(row["comment"]):
            replace_query.append(
                "CREATE OR REPLACE ICEBERG TABLE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            replace_query.append_nl(common_query)

            self.engine.execute_safe_ddl(replace_query)
            self.engine.execute_safe_ddl(
                "ALTER ICEBERG TABLE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": common_query.add_short_hash(bp.comment),
                },
            )

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP ICEBERG TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_unmanaged_iceberg_table_sql(self, bp: IcebergTableBlueprint):
        query = self.engine.query_builder()

        query.append_nl(
            "EXTERNAL_VOLUME = {external_volume:i}",
            {
                "external_volume": bp.external_volume,
            },
        )

        query.append_nl(
            "CATALOG = {catalog:i}",
            {
                "catalog": bp.catalog,
            },
        )

        if bp.catalog_table_name:
            query.append_nl(
                "CATALOG_TABLE_NAME = {catalog_table_name}",
                {
                    "catalog_table_name": bp.catalog_table_name,
                },
            )

        if bp.catalog_namespace:
            query.append_nl(
                "CATALOG_NAMESPACE = {catalog_namespace}",
                {
                    "catalog_namespace": bp.catalog_namespace,
                },
            )

        if bp.metadata_file_path:
            query.append_nl(
                "METADATA_FILE_PATH = {metadata_file_path}",
                {
                    "metadata_file_path": bp.metadata_file_path,
                },
            )

        if bp.base_location:
            query.append_nl(
                "BASE_LOCATION = {base_location}",
                {
                    "base_location": bp.base_location,
                },
            )

        if bp.replace_invalid_characters:
            query.append_nl("REPLACE_INVALID_CHARACTERS = TRUE")

        if bp.auto_refresh:
            query.append_nl("AUTO_REFRESH = TRUE")

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {"comment": bp.comment})

        return query



from snowddl.blueprint import NetworkRuleBlueprint, SchemaObjectIdent, Ident
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class NetworkRuleResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.NETWORK_RULE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW NETWORK RULES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "type": r["type"],
                "mode": r["mode"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(NetworkRuleBlueprint)

    def create_object(self, bp: NetworkRuleBlueprint):
        common_query = self._build_common_network_rule_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE NETWORK RULE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append(common_query)
        self.engine.execute_safe_ddl(create_query)

        return ResolveResult.CREATE

    def compare_object(self, bp: NetworkRuleBlueprint, row: dict):
        cur = self.engine.execute_meta(
            "DESC NETWORK RULE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        desc = cur.fetchone()

        if bp.type != row["type"] or bp.mode != row["mode"]:
            # Changing TYPE or MODE is only possible via full REPLACE
            self._drop_network_rule_refs(bp.full_name)

            common_query = self._build_common_network_rule_sql(bp)
            replace_query = self.engine.query_builder()

            replace_query.append(
                "CREATE OR REPLACE NETWORK RULE {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            replace_query.append(common_query)
            self.engine.execute_unsafe_ddl(replace_query)

            return ResolveResult.REPLACE

        result = ResolveResult.NOCHANGE

        if ",".join(bp.value_list) != desc["value_list"]:
            if bp.value_list:
                self.engine.execute_safe_ddl(
                    "ALTER NETWORK RULE {full_name:i} SET VALUE_LIST = ({value_list})",
                    {
                        "full_name": bp.full_name,
                        "value_list": bp.value_list,
                    },
                )
            else:
                self.engine.execute_safe_ddl(
                    "ALTER NETWORK RULE {full_name:i} UNSET VALUE_LIST",
                    {
                        "full_name": bp.full_name,
                    },
                )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER NETWORK RULE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        network_rule_name = SchemaObjectIdent("", row["database"], row["schema"], row["name"])

        self._drop_network_rule_refs(network_rule_name)
        self._drop_network_rule(network_rule_name)

        return ResolveResult.DROP

    def _build_common_network_rule_sql(self, bp: NetworkRuleBlueprint):
        query = self.engine.query_builder()

        query.append_nl(
            "TYPE = {type}",
            {
                "type": bp.type,
            },
        )

        query.append_nl(
            "MODE = {mode}",
            {
                "mode": bp.mode,
            },
        )

        if bp.value_list:
            query.append_nl(
                "VALUE_LIST = ({value_list})",
                {
                    "value_list": bp.value_list,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query

    def _drop_network_rule(self, network_rule_name: SchemaObjectIdent):
        self.engine.execute_unsafe_ddl(
            "DROP NETWORK RULE {full_name:i}",
            {
                "full_name": network_rule_name,
            },
        )

    def _drop_network_rule_refs(self, network_rule_name: SchemaObjectIdent):
        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.network_rule_references(network_rule_name => {network_rule_name}))",
            {
                "network_rule_name": network_rule_name,
            },
        )

        for r in cur:
            if r["CONTAINER_TYPE"] == "NETWORK_POLICY":
                self.engine.execute_unsafe_ddl(
                    "-- Required to drop NETWORK RULE\n"
                    "ALTER NETWORK POLICY {policy_name:i} REMOVE {list_type:r} = ({network_rule_name:i})",
                    {
                        "policy_name": Ident(r["CONTAINER_NAME"]),
                        "list_type": "ALLOWED_NETWORK_RULE_LIST" if r["ACTION_TYPE"] == "ALLOW" else "BLOCKED_NETWORK_RULE_LIST",
                        "network_rule_name": network_rule_name,
                    },
                )
            elif r["CONTAINER_TYPE"] == "INTEGRATION":
                # Since INTEGRATION does not support easy removal of individual rules, we have to reset all rules completely (for now)
                self.engine.execute_unsafe_ddl(
                    "-- Required to drop NETWORK RULE\n"
                    "ALTER EXTERNAL ACCESS INTEGRATION {integration_name:i} UNSET ALLOWED_NETWORK_RULES",
                    {
                        "integration_name": Ident(r["CONTAINER_NAME"]),
                        "network_rule_name": network_rule_name,
                    },
                )



from typing import TYPE_CHECKING

from snowddl.blueprint import Ident

if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class SchemaCache:
    def __init__(self, engine: "SnowDDLEngine"):
        self.engine = engine

        self.databases = {}
        self.schemas = {}

        self.reload()

    def reload(self):
        self.databases = {}
        self.schemas = {}

        cur = self.engine.execute_meta(
            "SHOW DATABASES LIKE {env_prefix:ls}",
            {
                "env_prefix": self.engine.config.env_prefix,
            },
        )

        for r in cur:
            # Skip databases created by other roles
            if r["owner"] != self.engine.context.current_role and not self.engine.settings.ignore_ownership:
                continue

            # Skip non-standard databases
            if r["kind"] != "STANDARD":
                continue

            # Skip databases not listed in settings explicitly
            if self.engine.settings.include_databases and Ident(r["name"]) not in self.engine.settings.include_databases:
                continue

            self.databases[r["name"]] = {
                "database": r["name"],
                "owner": r["owner"],
                "comment": r["comment"] if r["comment"] else None,
                "is_transient": "TRANSIENT" in r["options"],
                "retention_time": int(r["retention_time"]),
            }

        # Process schemas in parallel
        for database_schemas in self.engine.executor.map(self._get_database_schemas, self.databases):
            self.schemas.update(database_schemas)

    def _get_database_schemas(self, database_name):
        schemas = {}

        cur = self.engine.execute_meta(
            "SHOW SCHEMAS IN DATABASE {database:i}",
            {
                "database": database_name,
            },
        )

        for r in cur:
            # Skip INFORMATION_SCHEMA
            if r["name"] == "INFORMATION_SCHEMA":
                continue

            schemas[f"{r['database_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["name"],
                "owner": r["owner"],
                "comment": r["comment"] if r["comment"] else None,
                "is_transient": "TRANSIENT" in r["options"],
                "is_managed_access": "MANAGED ACCESS" in r["options"],
                "retention_time": int(r["retention_time"]) if r["retention_time"].isdigit() else 0,
            }

        return schemas



from snowddl.blueprint import TableBlueprint, SchemaObjectIdent
from snowddl.resolver.abc_schema_object_resolver import AbstractResolver, ResolveResult, ObjectType


class CloneTableResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.CLONE_TABLE

    def get_blueprints(self):
        return {}

    def get_existing_objects(self):
        databases_for_clone = self.get_databases_for_clone()
        schemas_for_clone = {}
        tables_for_clone = {}

        # Get schemas for clone in parallel
        for schema_objects in self.engine.executor.map(self.get_schemas_for_clone, databases_for_clone.values()):
            schemas_for_clone.update(schema_objects)

        # Get tables for clone in parallel
        for table_objects in self.engine.executor.map(self.get_tables_for_clone, schemas_for_clone.values()):
            tables_for_clone.update(table_objects)

        return tables_for_clone

    def get_databases_for_clone(self):
        databases_for_clone = {}
        clone_source_env_prefix = self.engine.settings.clone_source_env_prefix

        cur = self.engine.execute_meta("SHOW DATABASES")

        for r in cur:
            # Skip non-standard databases
            if r["kind"] != "STANDARD":
                continue

            src_database = str(r["name"])

            if clone_source_env_prefix:
                # Skip everything which does not start with source prefix
                if not src_database.startswith(clone_source_env_prefix):
                    continue

                dst_database = f"{self.config.env_prefix}{src_database.removeprefix(clone_source_env_prefix)}"
            else:
                dst_database = f"{self.config.env_prefix}{src_database}"

            # Skip every source database without destination database for cloning
            if dst_database not in self.engine.schema_cache.databases:
                continue

            databases_for_clone[dst_database] = {
                "src_database": src_database,
                "dst_database": dst_database,
            }

        return databases_for_clone

    def get_schemas_for_clone(self, database):
        schemas_for_clone = {}

        cur = self.engine.execute_meta(
            "SHOW SCHEMAS IN DATABASE {src_database:i}",
            {
                "src_database": database["src_database"],
            },
        )

        for r in cur:
            # Skip INFORMATION_SCHEMA
            if r["name"] == "INFORMATION_SCHEMA":
                continue

            dst_schema = f"{database['dst_database']}.{r['name']}"

            # Skip every source schema without destination schema for cloning
            if dst_schema not in self.engine.schema_cache.schemas:
                continue

            schemas_for_clone[dst_schema] = {
                "src_database": database["src_database"],
                "dst_database": database["dst_database"],
                "schema": r["name"],
            }

        return schemas_for_clone

    def get_tables_for_clone(self, schema):
        tables_for_clone = {}

        cur = self.engine.execute_meta(
            "SHOW TABLES IN SCHEMA {src_database:i}.{schema:i}",
            {
                "src_database": schema["src_database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Skip other table types
            if (
                r.get("is_external") == "Y"
                or r.get("is_event") == "Y"
                or r.get("is_hybrid") == "Y"
                or r.get("is_iceberg") == "Y"
                or r.get("is_dynamic") == "Y"
            ):
                continue

            tables_for_clone[f"{schema['dst_database']}.{r['schema_name']}.{r['name']}"] = {
                "src_database": schema["src_database"],
                "dst_database": schema["dst_database"],
                "schema": r["schema_name"],
                "name": r["name"],
                "is_transient": r["kind"] == "TRANSIENT",
            }

        return tables_for_clone

    def create_object(self, bp: TableBlueprint):
        return ResolveResult.NOCHANGE

    def compare_object(self, bp: TableBlueprint, row: dict):
        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        # Drop object is utilized since we have no blueprints for clones, but only existing objects
        # All such objects should be "dropped" from AbstractResolver perspective
        # TODO: find more elegant solution if cloning becomes popular

        query = self.engine.query_builder()
        query.append("CREATE")

        if row["is_transient"]:
            query.append("TRANSIENT")

        query.append(
            "TABLE IF NOT EXISTS {dst_database:i}.{schema:i}.{name:i}",
            {
                "dst_database": row["dst_database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        query.append_nl(
            "CLONE {src_database:i}.{schema:i}.{name:i}",
            {
                "src_database": row["src_database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        cur = self.engine.execute_clone(query)
        r = cur.fetchone()

        if str(r["status"]).endswith("successfully created."):
            self._drop_existing_policy_refs(
                ObjectType.TABLE, SchemaObjectIdent("", row["dst_database"], row["schema"], row["name"])
            )
            return ResolveResult.CREATE

        return ResolveResult.NOCHANGE

    def destroy(self):
        pass

    def _is_skipped(self):
        if not self.engine.settings.clone_table:
            return True

        if super()._is_skipped():
            return True

        return False

    def _drop_existing_policy_refs(self, object_type: ObjectType, object_name: SchemaObjectIdent):
        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.policy_references(ref_entity_domain => {object_type}, ref_entity_name => {object_name}))",
            {
                "object_type": object_type.singular_for_ref,
                "object_name": object_name,
            },
        )

        for r in cur:
            if r["POLICY_KIND"] == "AGGREGATION_POLICY":
                self.engine.execute_clone(
                    "ALTER {object_type:r} {object_name:i} UNSET AGGREGATION POLICY",
                    {
                        "object_type": object_type.singular_for_ref,
                        "object_name": object_name,
                    },
                )

            elif r["POLICY_KIND"] == "MASKING_POLICY":
                self.engine.execute_clone(
                    "ALTER {object_type:r} {object_name:i} MODIFY COLUMN {column:i} UNSET MASKING POLICY",
                    {
                        "object_type": object_type.singular_for_ref,
                        "object_name": object_name,
                        "column": r["REF_COLUMN_NAME"],
                    },
                )

            elif r["POLICY_KIND"] == "PROJECTION_POLICY":
                self.engine.execute_clone(
                    "ALTER {object_type:r} {object_name:i} MODIFY COLUMN {column:i} UNSET PROJECTION POLICY",
                    {
                        "object_type": object_type.singular_for_ref,
                        "object_name": object_name,
                        "column": r["REF_COLUMN_NAME"],
                    },
                )

            elif r["POLICY_KIND"] == "ROW_ACCESS_POLICY":
                policy_name = SchemaObjectIdent("", r["POLICY_DB"], r["POLICY_SCHEMA"], r["POLICY_NAME"])

                self.engine.execute_clone(
                    "ALTER {object_type:r} {object_name:i} DROP ROW ACCESS POLICY {policy_name:i}",
                    {
                        "object_type": object_type.singular_for_ref,
                        "object_name": object_name,
                        "policy_name": policy_name,
                    },
                )

            else:
                self.engine.logger.warning(
                    f"Detected unknown policy type [{r['POLICY_KIND']} attached to cloned object [{object_name}]"
                )



from re import compile

from snowddl.blueprint import ObjectType
from snowddl.converter.abc_converter import ConvertResult
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter
from snowddl.parser.sequence import sequence_json_schema
from snowddl.parser.table import table_json_schema


cluster_by_syntax_re = compile(r"^(\w+)?\((.*)\)$")
collate_type_syntax_re = compile(r"^(.*) COLLATE \'(.*)\'$")
identity_re = compile(r"^IDENTITY START (\d+) INCREMENT (\d+) (ORDER|NOORDER)$")


class TableConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Skip other table types
            if (
                r.get("is_external") == "Y"
                or r.get("is_event") == "Y"
                or r.get("is_hybrid") == "Y"
                or r.get("is_iceberg") == "Y"
                or r.get("is_dynamic") == "Y"
            ):
                continue

            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "is_transient": r["kind"] == "TRANSIENT",
                "cluster_by": r["cluster_by"] if r["cluster_by"] else None,
                "change_tracking": bool(r["change_tracking"] == "ON"),
                "search_optimization": bool(r.get("search_optimization") == "ON"),
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def dump_object(self, row):
        cols, identities = self._get_columns(row)

        data = {"columns": cols}

        if row["is_transient"]:
            data["is_transient"] = True

        if row["cluster_by"]:
            data["cluster_by"] = cluster_by_syntax_re.sub(r"\2", row["cluster_by"]).split(", ")

        if row["change_tracking"]:
            data["change_tracking"] = True

        if row["search_optimization"]:
            data["search_optimization"] = True

        data["comment"] = row["comment"]

        data["primary_key"] = self._get_primary_key(row)
        data["unique_keys"] = self._get_unique_keys(row)
        data["foreign_keys"] = self._get_foreign_keys(row)

        object_path = (
            self.base_path / self._normalise_name_with_prefix(row["database"]) / self._normalise_name(row["schema"]) / "table"
        )

        self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, table_json_schema)

        for sequence_name, sequence_data in identities.items():
            object_path = (
                self.base_path
                / self._normalise_name_with_prefix(row["database"])
                / self._normalise_name(row["schema"])
                / "sequence"
            )
            object_path.mkdir(mode=0o755, parents=True, exist_ok=True)

            self._dump_file(object_path / f"{self._normalise_name(sequence_name)}.yaml", sequence_data, sequence_json_schema)

        return ConvertResult.DUMP

    def _get_columns(self, row):
        cols = {}
        identities = {}

        cur = self.engine.execute_meta(
            "DESC TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        for c in cur:
            m = collate_type_syntax_re.match(c["type"])

            if m:
                col = {
                    "type": m.group(1),
                    "collate": m.group(2),
                }
            else:
                col = {"type": c["type"]}

            if c["null?"] == "N":
                col["type"] = f"{col['type']} NOT NULL"

            if c["default"]:
                i = identity_re.match(c["default"])

                if i:
                    sequence_name = self._get_auto_sequence_name(row["name"], c["name"])
                    col["default_sequence"] = sequence_name

                    identities[sequence_name] = {
                        "start": int(i.group(1)),
                        "interval": int(i.group(2)),
                        "is_ordered": i.group(3) == "ORDER",
                    }
                elif str(c["default"]).upper().endswith(".NEXTVAL"):
                    col["default_sequence"] = self._normalise_name_with_prefix(str(c["default"])[:-8])
                else:
                    col["default"] = str(c["default"])

            if c["expression"]:
                col["expression"] = c["expression"]

            if c["comment"]:
                col["comment"] = c["comment"]

            cols[self._normalise_name(c["name"])] = col

        return cols, identities

    def _get_primary_key(self, row):
        constraints = {}

        cur = self.engine.execute_meta(
            "SHOW PRIMARY KEYS IN TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        for r in cur:
            if r["constraint_name"] not in constraints:
                constraints[r["constraint_name"]] = {}

            constraints[r["constraint_name"]][r["key_sequence"]] = r["column_name"]

        if not constraints:
            return None

        # It is possible to have multiple PRIMARY KEYS in some cases
        # Return only the first key
        for pk in constraints.values():
            return [self._normalise_name(pk[seq]) for seq in sorted(pk)]

    def _get_unique_keys(self, row):
        constraints = {}
        unique_keys = []

        cur = self.engine.execute_meta(
            "SHOW UNIQUE KEYS IN TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        for r in cur:
            if r["constraint_name"] not in constraints:
                constraints[r["constraint_name"]] = {}

            constraints[r["constraint_name"]][r["key_sequence"]] = r["column_name"]

        if not constraints:
            return None

        for uq in constraints.values():
            unique_keys.append([self._normalise_name(uq[seq]) for seq in sorted(uq)])

        return unique_keys

    def _get_foreign_keys(self, row):
        constraints = {}
        foreign_keys = []

        cur = self.engine.execute_meta(
            "SHOW IMPORTED KEYS IN TABLE {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        for r in cur:
            if r["fk_name"] not in constraints:
                constraints[r["fk_name"]] = {
                    "columns": {},
                    "ref_table": f"{r['pk_database_name']}.{r['pk_schema_name']}.{r['pk_table_name']}",
                    "ref_columns": {},
                }

            constraints[r["fk_name"]]["columns"][r["key_sequence"]] = r["fk_column_name"]
            constraints[r["fk_name"]]["ref_columns"][r["key_sequence"]] = r["pk_column_name"]

        if not constraints:
            return None

        for fk in constraints.values():
            foreign_keys.append(
                {
                    "columns": [self._normalise_name(fk["columns"][seq]) for seq in sorted(fk["columns"])],
                    "ref_table": self._normalise_name_with_prefix(fk["ref_table"]),
                    "ref_columns": [self._normalise_name(fk["ref_columns"][seq]) for seq in sorted(fk["ref_columns"])],
                }
            )

        return foreign_keys

    def _get_auto_sequence_name(self, table_name, column_name):
        return self._normalise_name_with_prefix(f"{table_name}_{column_name}_seq")



from snowddl.blueprint import (
    BusinessRoleBlueprint,
    DatabaseBlueprint,
    Grant,
    Ident,
    RoleBlueprint,
    SchemaBlueprint,
    build_role_ident,
)
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class ShareAccessRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.SHARE_ACCESS_ROLE_SUFFIX

    def get_blueprints(self):
        blueprints = []
        unique_share_names = set()

        for database_bp in self.config.get_blueprints_by_type(DatabaseBlueprint).values():
            for share_name in database_bp.owner_share_read:
                if isinstance(share_name, Ident):
                    unique_share_names.add(share_name)

        for schema_bp in self.config.get_blueprints_by_type(SchemaBlueprint).values():
            for share_name in schema_bp.owner_share_read:
                if isinstance(share_name, Ident):
                    unique_share_names.add(share_name)

        for business_role_bp in self.config.get_blueprints_by_type(BusinessRoleBlueprint).values():
            for share_name in business_role_bp.share_read:
                if isinstance(share_name, Ident):
                    unique_share_names.add(share_name)

        for share_name in unique_share_names:
            blueprints.append(self.get_blueprint_share_role(share_name))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_share_role(self, share_name: Ident):
        grants = [
            Grant(
                privilege="IMPORTED PRIVILEGES",
                on=ObjectType.DATABASE,
                name=share_name,
            ),
        ]

        return RoleBlueprint(
            full_name=build_role_ident(self.config.env_prefix, share_name, self.config.SHARE_ACCESS_ROLE_SUFFIX),
            grants=grants,
        )

    def get_existing_role_grants(self, role_name):
        grants = []
        share_database_name = Ident(role_name[len(self.config.env_prefix) : -len(self.get_role_suffix()) - 2])

        # Grant is being checked in reverse for performance reasons
        # Traditional SHOW GRANTS on role with IMPORTED PRIVILEGES might be very slow
        cur = self.engine.execute_meta(
            "SHOW GRANTS ON DATABASE {share_database_name:i}",
            {
                "share_database_name": share_database_name,
            },
        )

        for r in cur:
            if r["privilege"] == "USAGE" and r["granted_to"] == "ROLE" and r["grantee_name"] == role_name:
                # There is no way to detect specific grant of IMPORTED PRIVILEGES
                # This grant has to be emulated until we find a better option
                grants.append(
                    Grant(
                        privilege="IMPORTED PRIVILEGES",
                        on=ObjectType.DATABASE,
                        name=share_database_name,
                    )
                )

        return role_name, grants, [], []



from snowddl.blueprint import EventTableBlueprint, StreamBlueprint, TableBlueprint, ViewBlueprint
from snowddl.validator.abc_validator import AbstractValidator


class StreamValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StreamBlueprint)

    def validate_blueprint(self, bp: StreamBlueprint):
        self._validate_dependency(bp)

    def _validate_dependency(self, bp: StreamBlueprint):
        if bp.object_name not in self.config.get_blueprints_by_type(bp.object_type.blueprint_cls):
            raise ValueError(
                f"Stream [{bp.full_name}] depends on {bp.object_type.name} " f"[{bp.object_name}] which does not exist in config"
            )

        target_bp = self.config.get_blueprints_by_type(bp.object_type.blueprint_cls)[str(bp.object_name)]

        if isinstance(target_bp, (EventTableBlueprint, TableBlueprint, ViewBlueprint)):
            if not target_bp.change_tracking:
                raise ValueError(
                    f"Stream [{bp.full_name}] depends on {bp.object_type.name} "
                    f"[{bp.object_name}], but change_tracking is not enabled on it"
                )



from abc import ABC, abstractmethod
from logging import getLogger, NullHandler
from traceback import TracebackException
from typing import Dict

from snowddl.blueprint import AbstractBlueprint
from snowddl.config import SnowDDLConfig


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class AbstractValidator(ABC):
    def __init__(self, config: SnowDDLConfig):
        self.config = config
        self.env_prefix = config.env_prefix

        self.logger = logger
        self.errors: Dict[str, Exception] = {}

    def validate(self):
        for full_name, bp in self.get_blueprints().items():
            try:
                self.validate_blueprint(bp)
            except Exception as exc:
                traceback = "".join(TracebackException.from_exception(exc).format())
                logger.warning(f"Failed to validate config for [{bp.__class__.__name__}] with name [{full_name}]\n{traceback}")
                self.errors[full_name] = exc

    @abstractmethod
    def validate_blueprint(self, bp: AbstractBlueprint):
        pass

    @abstractmethod
    def get_blueprints(self) -> Dict[str, AbstractBlueprint]:
        pass



from snowddl.blueprint import (
    AccountObjectIdent,
    BusinessRoleBlueprint,
    Ident,
    IdentPattern,
    build_role_ident,
    build_application_role_ident,
    build_share_read_ident,
)
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
business_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "database_owner": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "database_write": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "database_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_owner": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_write": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "schema_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "share_read": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "warehouse_usage": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "warehouse_monitor": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "technical_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "application_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "tech_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "global_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "additionalProperties": False,
    }
}
# fmt: on


class BusinessRoleParser(AbstractParser):
    def load_blueprints(self):
        self.parse_multi_entity_file("business_role", business_role_json_schema, self.process_business_role)

    def process_business_role(self, business_role_name, business_role_params):
        application_roles = []
        technical_roles = []

        for application_role_name in business_role_params.get("application_roles", []):
            application_roles.append(build_application_role_ident(application_role_name))

        for technical_role_name in business_role_params.get("technical_roles", []) + business_role_params.get("tech_roles", []):
            technical_roles.append(build_role_ident(self.env_prefix, technical_role_name, self.config.TECHNICAL_ROLE_SUFFIX))

        # fmt: off
        bp = BusinessRoleBlueprint(
            full_name=build_role_ident(self.env_prefix, business_role_name, self.config.BUSINESS_ROLE_SUFFIX),
            database_owner=[IdentPattern(p) for p in business_role_params.get("database_owner", [])],
            database_write=[IdentPattern(p) for p in business_role_params.get("database_write", [])],
            database_read=[IdentPattern(p) for p in business_role_params.get("database_read", [])],
            schema_owner=[IdentPattern(p) for p in business_role_params.get("schema_owner", [])],
            schema_write=[IdentPattern(p) for p in business_role_params.get("schema_write", [])],
            schema_read=[IdentPattern(p) for p in business_role_params.get("schema_read", [])],
            share_read=[build_share_read_ident(share_name) for share_name in business_role_params.get("share_read", [])],
            warehouse_usage=[AccountObjectIdent(self.env_prefix, warehouse_name) for warehouse_name in business_role_params.get("warehouse_usage", [])],
            warehouse_monitor=[AccountObjectIdent(self.env_prefix, warehouse_name) for warehouse_name in business_role_params.get("warehouse_monitor", [])],
            application_roles=application_roles,
            technical_roles=technical_roles,
            global_roles=[Ident(global_role_name) for global_role_name in business_role_params.get("global_roles", [])],
            comment=business_role_params.get("comment"),
        )
        # fmt: on

        self.config.add_blueprint(bp)



from pathlib import Path

try:
    from sqlglot import Tokenizer, TokenType
except ImportError:
    Tokenizer = None
    TokenType = None

from snowddl.blueprint import ObjectType
from snowddl.converter.abc_converter import ConvertResult
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter
from snowddl.converter._yaml import YamlLiteralStr, YamlIncludeStr
from snowddl.parser.view import view_json_schema


class ViewConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW VIEWS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            if r["is_materialized"] == "true":
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "text": r["text"],
                "is_secure": r["is_secure"] == "true",
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def dump_object(self, row):
        object_path = (
            self.base_path / self._normalise_name_with_prefix(row["database"]) / self._normalise_name(row["schema"]) / "view"
        )

        data = {
            "columns": self._get_columns(row),
            "text": self._get_text_or_include(object_path, row),
            "comment": row["comment"],
        }

        if data:
            self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, view_json_schema)
            return ConvertResult.DUMP

        return ConvertResult.EMPTY

    def _get_columns(self, row):
        cols = {}

        cur = self.engine.execute_meta(
            "DESC VIEW {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        for c in cur:
            cols[self._normalise_name(c["name"])] = c["comment"] if c["comment"] else None

        return cols

    def _get_text_or_include(self, object_path: Path, row: dict):
        view_text = self._extract_view_text_after_as(row)
        view_text = view_text.rstrip(";").strip(" \n\r\t")

        # Remove trailing spaces from each line to prevent output formatting issues
        # https://github.com/yaml/pyyaml/issues/411
        view_text = "\n".join(line.rstrip(" ") for line in view_text.split("\n"))

        if self.engine.settings.convert_view_text_to_file:
            file_name = f"{self._normalise_name(row['name'])}.sql"
            self._dump_code(object_path / "sql" / file_name, view_text)

            return YamlIncludeStr(f"sql/{file_name}")

        return YamlLiteralStr(view_text)

    def _extract_view_text_after_as(self, row: dict):
        tokenizer = Tokenizer()

        for token in tokenizer.tokenize(row["text"]):
            if token.token_type == TokenType.ALIAS:
                return row["text"][token.end+2:]

        raise ValueError(f"Could not extract view text after 'AS' keyword using sqlglot tokenizer")

    def _check_optional_dependencies(self):
        if Tokenizer is None:
            raise ImportError("Package [sqlglot] is required to parse VIEW text")



from snowddl.blueprint import DynamicTableBlueprint
from snowddl.validator.abc_validator import AbstractValidator


class DynamicTableValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DynamicTableBlueprint)

    def validate_blueprint(self, bp: DynamicTableBlueprint):
        self._validate_depends_on(bp)

    def _validate_depends_on(self, bp: DynamicTableBlueprint):
        for depends_on_name in bp.depends_on:
            if depends_on_name == bp.full_name:
                raise ValueError(f"Dynamic table [{bp.full_name}] cannot depend on itself")

            if depends_on_name not in self.config.get_blueprints_by_type(DynamicTableBlueprint):
                raise ValueError(
                    f"Dynamic table [{bp.full_name}] depends on another dynamic table "
                    f"[{depends_on_name}] which does not exist in config"
                )



from typing import List

from snowddl.blueprint import DatabaseRoleBlueprint, Grant, GrantPattern, build_grant_name_ident
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class DatabaseRoleResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.DATABASE_ROLE

    def get_existing_objects(self):
        existing_objects = {}

        # Process schemas in parallel
        for database_objects in self.engine.executor.map(
            self.get_existing_objects_in_database, self.engine.schema_cache.databases.values()
        ):
            existing_objects.update(database_objects)

        return existing_objects

    def get_existing_objects_in_database(self, database: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW DATABASE ROLES IN DATABASE {database:i}",
            {
                "database": database["database"],
            },
        )

        for r in cur:
            full_name = f"{database['database']}.{r['name']}"
            existing_objects[full_name] = {
                "database": database["database"],
                "name": r["name"],
                "owner": r["owner"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(DatabaseRoleBlueprint)

    def create_object(self, bp: DatabaseRoleBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE DATABASE ROLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_safe_ddl(query)

        for bp_grant in self.expand_grant_patterns_to_grants(bp.grant_patterns):
            self.create_grant(bp.full_name, bp_grant)

        return ResolveResult.CREATE

    def compare_object(self, bp: DatabaseRoleBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        bp_grants = self.expand_grant_patterns_to_grants(bp.grant_patterns)
        existing_grants = self.get_existing_grants(bp.full_name)

        for bp_grant in bp_grants:
            if bp_grant not in existing_grants:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        for ex_grant in existing_grants:
            if not any(grant_pattern.is_matching_grant(ex_grant) for grant_pattern in bp.grant_patterns):
                self.drop_grant(bp.full_name, ex_grant)
                result = ResolveResult.GRANT

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER DATABASE ROLE {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP DATABASE ROLE {database:i}.{name:i}",
            {
                "database": row["database"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def create_grant(self, role_name, grant: Grant):
        self.engine.execute_unsafe_ddl(
            "GRANT {privilege:r} ON {on:r} {name:i} TO DATABASE ROLE {role_name:i}",
            {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "role_name": role_name,
            },
        )

    def drop_grant(self, role_name, grant: Grant):
        self.engine.execute_unsafe_ddl(
            "REVOKE {privilege:r} ON {on:r} {name:i} FROM DATABASE ROLE {role_name:i}",
            {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "role_name": role_name,
            },
        )

    def get_existing_grants(self, role_name):
        grants = []

        cur = self.engine.execute_meta(
            "SHOW GRANTS TO DATABASE ROLE {role_name:i}",
            {
                "role_name": role_name,
            },
        )

        for r in cur:
            grants.append(
                Grant(
                    privilege=r["privilege"],
                    on=ObjectType[r["granted_on"]],
                    name=build_grant_name_ident(self.config.env_prefix, r["name"], ObjectType[r["granted_on"]]),
                )
            )

        return grants

    def expand_grant_patterns_to_grants(self, grant_patterns: List[GrantPattern]):
        grants = []

        for grant_pattern in grant_patterns:
            blueprints = self.config.get_blueprints_by_type_and_pattern(grant_pattern.on.blueprint_cls, grant_pattern.pattern)

            for obj_bp in blueprints.values():
                grants.append(
                    Grant(
                        privilege=grant_pattern.privilege,
                        on=grant_pattern.on,
                        name=obj_bp.full_name,
                    ),
                )

        return grants



from snowddl.blueprint import ProjectionPolicyBlueprint, ObjectType, Edition, SchemaObjectIdent
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class ProjectionPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True
    skip_min_edition = Edition.ENTERPRISE

    def get_object_type(self) -> ObjectType:
        return ObjectType.PROJECTION_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW PROJECTION POLICIES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ProjectionPolicyBlueprint)

    def create_object(self, bp: ProjectionPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: ProjectionPolicyBlueprint, row: dict):
        cur = self.engine.execute_meta(
            "DESC PROJECTION POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        r = cur.fetchone()
        result = ResolveResult.NOCHANGE

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        if r["body"] != bp.body:
            self.engine.execute_unsafe_ddl(
                "ALTER PROJECTION POLICY {full_name:i} SET BODY -> {body:r}",
                {
                    "full_name": bp.full_name,
                    "body": bp.body,
                },
                condition=self.engine.settings.execute_projection_policy,
            )

            result = ResolveResult.ALTER

        if row["comment"] != bp.comment:
            self.engine.execute_unsafe_ddl(
                "ALTER PROJECTION POLICY {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
                condition=self.engine.settings.execute_projection_policy,
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_policy_refs(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))
        self._drop_policy(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))

        return ResolveResult.DROP

    def _create_policy(self, bp: ProjectionPolicyBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE PROJECTION POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append_nl(
            "AS () RETURNS PROJECTION_CONSTRAINT -> {body:r}",
            {
                "body": bp.body,
            },
        )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_projection_policy)

    def _drop_policy(self, policy: SchemaObjectIdent):
        self.engine.execute_unsafe_ddl(
            "DROP PROJECTION POLICY {full_name:i}",
            {"full_name": policy},
            condition=self.engine.settings.execute_projection_policy,
        )

    def _apply_policy_refs(self, bp: ProjectionPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.full_name)
        applied_change = False

        for ref in bp.references:
            ref_key = f"{ref.object_type.name}|{ref.object_name}|{ref.column}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                del existing_policy_refs[ref_key]
                continue

            # Apply new policy or replace existing policy
            self.engine.execute_unsafe_ddl(
                "ALTER {object_type:r} {object_name:i} MODIFY COLUMN {column:i} SET PROJECTION POLICY {policy_name:i} FORCE",
                {
                    "object_type": ref.object_type.singular_for_ref,
                    "object_name": ref.object_name,
                    "column": ref.column,
                    "policy_name": bp.full_name,
                },
                condition=self.engine.settings.execute_projection_policy,
            )

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            # TODO: consider use case when object switches to another projection policy resolved in parallel
            self.engine.execute_unsafe_ddl(
                "ALTER {object_type:r} {database:i}.{schema:i}.{name:i} MODIFY COLUMN {column:i} UNSET PROJECTION POLICY",
                {
                    "object_type": ObjectType[existing_ref["object_type"]].singular_for_ref,
                    "database": existing_ref["database"],
                    "schema": existing_ref["schema"],
                    "column": existing_ref["column"],
                    "name": existing_ref["name"],
                },
                condition=self.engine.settings.execute_projection_policy,
            )

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = self._get_existing_policy_refs(policy_name)

        for existing_ref in existing_policy_refs.values():
            # TODO: consider use case when object switches to another projection policy resolved in parallel
            self.engine.execute_unsafe_ddl(
                "ALTER {object_type:r} {database:i}.{schema:i}.{name:i} MODIFY COLUMN {column:i} UNSET PROJECTION POLICY",
                {
                    "object_type": ObjectType[existing_ref["object_type"]].singular_for_ref,
                    "database": existing_ref["database"],
                    "schema": existing_ref["schema"],
                    "column": existing_ref["column"],
                    "name": existing_ref["name"],
                },
                condition=self.engine.settings.execute_projection_policy,
            )

    def _get_existing_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = {}

        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.policy_references(policy_name => {policy_name}))",
            {
                "policy_name": policy_name,
            },
        )

        for r in cur:
            ref_key = (
                f"{r['REF_ENTITY_DOMAIN']}"
                f"|{r['REF_DATABASE_NAME']}.{r['REF_SCHEMA_NAME']}.{r['REF_ENTITY_NAME']}"
                f"|{r['REF_COLUMN_NAME']}"
            )

            existing_policy_refs[ref_key] = {
                "object_type": r["REF_ENTITY_DOMAIN"],
                "database": r["REF_DATABASE_NAME"],
                "schema": r["REF_SCHEMA_NAME"],
                "name": r["REF_ENTITY_NAME"],
                "column": r["REF_COLUMN_NAME"],
            }

        return existing_policy_refs



from typing import List, Optional

from .ident import Ident, SchemaObjectIdent
from ..model import BaseModelWithConfig


class SemanticViewTable(BaseModelWithConfig):
    table_alias: Optional[Ident] = None
    table_name: SchemaObjectIdent
    primary_key: Optional[List[Ident]] = None
    unique: Optional[List[List[Ident]]] = None
    with_synonyms: Optional[List[str]] = None
    comment: Optional[str] = None


class SemanticViewRelationship(BaseModelWithConfig):
    relationship_identifier: Optional[Ident] = None
    table_alias: Ident
    columns: List[Ident]
    ref_table_alias: Ident
    ref_columns: List[Ident]


class SemanticViewExpression(BaseModelWithConfig):
    table_alias: Ident
    name: Ident
    sql: str
    with_synonyms: Optional[List[str]] = None
    comment: Optional[str] = None



from snowddl.blueprint import AuthenticationPolicyBlueprint, ObjectType, SchemaObjectIdent
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class AuthenticationPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.AUTHENTICATION_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW AUTHENTICATION POLICIES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(AuthenticationPolicyBlueprint)

    def create_object(self, bp: AuthenticationPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: AuthenticationPolicyBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._compare_policy(bp, row):
            result = ResolveResult.ALTER

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_policy_refs(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))
        self._drop_policy(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))

        return ResolveResult.DROP

    def _create_policy(self, bp: AuthenticationPolicyBlueprint):
        common_query = self._build_common_authentication_policy_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE AUTHENTICATION POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append(common_query)
        self.engine.execute_unsafe_ddl(create_query, condition=self.engine.settings.execute_authentication_policy)

        self.engine.execute_unsafe_ddl(
            "COMMENT ON AUTHENTICATION POLICY {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            },
            condition=self.engine.settings.execute_authentication_policy
        )

    def _compare_policy(self, bp: AuthenticationPolicyBlueprint, row: dict):
        common_query = self._build_common_authentication_policy_sql(bp)

        if not common_query.compare_short_hash(row["comment"]):
            alter_query = self.engine.query_builder()

            alter_query.append(
                "CREATE OR ALTER AUTHENTICATION POLICY {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

            alter_query.append(common_query)
            self.engine.execute_unsafe_ddl(alter_query, condition=self.engine.settings.execute_authentication_policy)

            self.engine.execute_safe_ddl(
                "COMMENT ON AUTHENTICATION POLICY {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": common_query.add_short_hash(bp.comment),
                },
                condition=self.engine.settings.execute_authentication_policy
            )

            return True

        return False

    def _build_common_authentication_policy_sql(self, bp: AuthenticationPolicyBlueprint):
        query = self.engine.query_builder()

        if bp.authentication_methods:
            query.append_nl(
                "AUTHENTICATION_METHODS = ({authentication_methods})",
                {
                    "authentication_methods": bp.authentication_methods,
                },
            )

        if bp.mfa_authentication_methods:
            query.append_nl(
                "MFA_AUTHENTICATION_METHODS = ({mfa_authentication_methods})",
                {
                    "mfa_authentication_methods": bp.mfa_authentication_methods,
                },
            )

        if bp.mfa_enrollment:
            query.append_nl(
                "MFA_ENROLLMENT = {mfa_enrollment}",
                {
                    "mfa_enrollment": bp.mfa_enrollment,
                },
            )

        if bp.mfa_policy:
            query.append_nl("MFA_POLICY = (")

            for param_name, param_value in bp.mfa_policy.items():
                query.append(
                    "{param_name:r} = {param_value:dp}",
                    {
                        "param_name": param_name,
                        "param_value": param_value,
                    }
                )

            query.append(")")

        if bp.client_types:
            query.append_nl(
                "CLIENT_TYPES = ({client_types})",
                {
                    "client_types": bp.client_types,
                },
            )

        if bp.client_policy:
            query.append_nl("CLIENT_POLICY = (")

            for client_name, client_params in bp.client_policy.items():
                query.append(
                    "{client_name:r} = (",
                    {
                        "client_name": client_name,
                    }
                )

                for param_name, param_value in client_params.items():
                    query.append(
                        "{param_name:r} = {param_value:dp}",
                        {
                            "param_name": param_name,
                            "param_value": param_value,
                        }
                    )

                query.append(")")

            query.append(")")

        if bp.security_integrations:
            query.append_nl(
                "SECURITY_INTEGRATIONS = ({security_integrations})",
                {
                    "security_integrations": bp.security_integrations,
                },
            )

        if bp.pat_policy:
            query.append_nl("PAT_POLICY = (")

            for param_name, param_value in bp.pat_policy.items():
                query.append(
                    "{param_name:r} = {param_value:dp}",
                    {
                        "param_name": param_name,
                        "param_value": param_value,
                    }
                )

            query.append(")")

        if bp.workload_identity_policy:
            query.append_nl("WORKLOAD_IDENTITY_POLICY = (")

            for param_name, param_value in bp.workload_identity_policy.items():
                query.append(
                    "{param_name:r} = {param_value:dp}",
                    {
                        "param_name": param_name,
                        "param_value": param_value,
                    }
                )

            query.append(")")

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query

    def _drop_policy(self, policy: SchemaObjectIdent):
        self.engine.execute_unsafe_ddl(
            "DROP AUTHENTICATION POLICY {full_name:i}",
            {"full_name": policy},
            condition=self.engine.settings.execute_authentication_policy,
        )

    def _apply_policy_refs(self, bp: AuthenticationPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.full_name)
        applied_change = False

        for ref in bp.references:
            if ref.object_type == ObjectType.ACCOUNT:
                ref_key = ref.object_type.name
            else:
                ref_key = f"{ref.object_type.name}|{ref.object_name}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                del existing_policy_refs[ref_key]
                continue

            if ref.object_type == ObjectType.ACCOUNT:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT SET AUTHENTICATION POLICY {policy_name:i} FORCE",
                    {
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_authentication_policy
                    and self.engine.settings.execute_account_level_policy,
                )
            else:
                # Apply new policy for USER (and other object types in future?)
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} SET AUTHENTICATION POLICY {policy_name:i} FORCE",
                    {
                        "object_type": ref.object_type.name,
                        "object_name": ref.object_name,
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_authentication_policy,
                )

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            if existing_ref["object_type"] == ObjectType.ACCOUNT.name:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT UNSET AUTHENTICATION POLICY",
                    condition=self.engine.settings.execute_authentication_policy
                    and self.engine.settings.execute_account_level_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} UNSET AUTHENTICATION POLICY",
                    {
                        "object_type": existing_ref["object_type"],
                        "object_name": existing_ref["name"],
                    },
                    condition=self.engine.settings.execute_authentication_policy,
                )

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = self._get_existing_policy_refs(policy_name)

        for existing_ref in existing_policy_refs.values():
            if existing_ref["object_type"] == ObjectType.ACCOUNT.name:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT UNSET AUTHENTICATION POLICY",
                    condition=self.engine.settings.execute_authentication_policy
                    and self.engine.settings.execute_account_level_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} UNSET AUTHENTICATION POLICY",
                    {
                        "object_type": existing_ref["object_type"],
                        "object_name": existing_ref["name"],
                    },
                    condition=self.engine.settings.execute_authentication_policy,
                )

    def _get_existing_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = {}

        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.policy_references(policy_name => {policy_name}))",
            {
                "policy_name": policy_name,
            },
        )

        for r in cur:
            if r["REF_ENTITY_DOMAIN"] == ObjectType.ACCOUNT.name:
                ref_key = r["REF_ENTITY_DOMAIN"]
            else:
                ref_key = f"{r['REF_ENTITY_DOMAIN']}|{r['REF_ENTITY_NAME']}"

            existing_policy_refs[ref_key] = {
                "object_type": r["REF_ENTITY_DOMAIN"],
                "name": r["REF_ENTITY_NAME"],
            }

        return existing_policy_refs



from snowddl.blueprint import PipeBlueprint, Ident, SchemaObjectIdent, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
pipe_json_schema = {
    "type": "object",
    "properties": {
        "copy": {
            "type": "object",
            "properties": {
                "table": {
                    "type": "string"
                },
                "stage": {
                    "type": "string"
                },
                "path": {
                    "type": "string"
                },
                "pattern": {
                    "type": "string"
                },
                "file_format": {
                    "type": "string"
                },
                "match_by_column_name": {
                    "type": "string"
                },
                "include_metadata": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                },
                "transform": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                },
                "options": {
                    "type": "object",
                    "additionalProperties": {
                        "type": ["array", "boolean", "number", "string"]
                    }
                }
            },
            "required": ["table", "stage"],
            "additionalProperties": False
        },
        "auto_ingest": {
            "type": "boolean"
        },
        "aws_sns_topic": {
            "type": "string"
        },
        "integration": {
            "type": "string"
        },
        "error_integration": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["copy", "auto_ingest"],
    "additionalProperties": False
}
# fmt: on


class PipeParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("pipe", pipe_json_schema, self.process_pipe)

    def process_pipe(self, f: ParsedFile):
        file_format = None
        include_metadata = None

        copy = f.params["copy"]

        if copy.get("file_format"):
            file_format = build_schema_object_ident(self.env_prefix, copy.get("file_format"), f.database, f.schema)

        if copy.get("include_metadata"):
            include_metadata = {Ident(k): Ident(v) for k, v in copy.get("include_metadata").items()}

        bp = PipeBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            auto_ingest=f.params["auto_ingest"],
            copy_table_name=build_schema_object_ident(self.env_prefix, copy["table"], f.database, f.schema),
            copy_stage_name=build_schema_object_ident(self.env_prefix, copy["stage"], f.database, f.schema),
            copy_path=copy.get("path"),
            copy_pattern=copy.get("pattern"),
            copy_transform=self.normalise_params_dict(copy.get("transform")),
            copy_file_format=file_format,
            copy_match_by_column_name=copy.get("match_by_column_name"),
            copy_include_metadata=include_metadata,
            copy_options=self.normalise_params_dict(copy.get("options")),
            aws_sns_topic=f.params.get("aws_sns_topic"),
            integration=Ident(f.params["integration"]) if f.params.get("integration") else None,
            error_integration=Ident(f.params["error_integration"]) if f.params.get("error_integration") else None,
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import BackupPolicyBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
backup_policy_json_schema = {
    "type": "object",
    "properties": {
        "schedule": {
            "type": "string"
        },
        "expire_after_days": {
            "type": "integer"
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class BackupPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("backup_policy", backup_policy_json_schema, self.process_backup_policy)

    def process_backup_policy(self, f: ParsedFile):
        bp = BackupPolicyBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            schedule=f.params.get("schedule"),
            expire_after_days=f.params.get("expire_after_days"),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import ObjectType
from snowddl.converter.abc_converter import AbstractConverter, ConvertResult
from snowddl.parser.schema import schema_json_schema


class SchemaConverter(AbstractConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.SCHEMA

    def get_existing_objects(self):
        return self.engine.schema_cache.schemas

    def dump_object(self, row):
        data = {}

        if row["is_transient"]:
            data["is_transient"] = True

        if row["comment"]:
            data["comment"] = row["comment"]

        object_path = self.base_path / self._normalise_name_with_prefix(row["database"]) / self._normalise_name(row["schema"])

        self._dump_file(object_path / "params.yaml", data, schema_json_schema)
        return ConvertResult.DUMP



from itertools import islice
from re import compile

from snowddl.blueprint import (
    Ident,
    TableBlueprint,
    TableColumn,
    DataType,
    BaseDataType,
    SchemaObjectIdent,
    SearchOptimizationItem,
)
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType

cluster_by_syntax_re = compile(r"^(\w+)?\((.*)\)$")
collate_type_syntax_re = compile(r"^(.*) COLLATE \'(.*)\'$")


class TableResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Skip other table types
            if (
                r.get("is_external") == "Y"
                or r.get("is_event") == "Y"
                or r.get("is_hybrid") == "Y"
                or r.get("is_iceberg") == "Y"
                or r.get("is_dynamic") == "Y"
            ):
                continue

            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "is_transient": r["kind"] == "TRANSIENT",
                "retention_time": int(r["retention_time"]),
                "cluster_by": r["cluster_by"] if r["cluster_by"] else None,
                "change_tracking": bool(r["change_tracking"] == "ON"),
                "search_optimization": bool(r.get("search_optimization") == "ON"),
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TableBlueprint)

    def create_object(self, bp: TableBlueprint):
        query = self._build_create_table(bp)
        self.engine.execute_safe_ddl(query)

        self._create_search_optimization(bp)

        return ResolveResult.CREATE

    def compare_object(self, bp: TableBlueprint, row: dict):
        safe_alters = []
        unsafe_alters = []

        replace_reasons = []
        replace_notices = []

        bp_cols = {str(c.name): c for c in bp.columns}
        snow_cols = self._get_existing_columns(bp)

        dropping_col_names = []
        remaining_col_names = list(snow_cols.keys())

        for col_name, snow_c in snow_cols.items():
            # Drop columns which do not exist in blueprint
            if col_name not in bp_cols:
                unsafe_alters.append(
                    self.engine.format(
                        "DROP COLUMN {col_name:i}",
                        {
                            "col_name": col_name,
                        },
                    )
                )

                dropping_col_names.append(col_name)
                remaining_col_names.remove(col_name)

                replace_notices.append(f"Column {col_name} is about to be dropped")
                continue

            bp_c = bp_cols[col_name]

            # Set or drop NOT NULL constraint
            if snow_c.not_null and not bp_c.not_null:
                unsafe_alters.append(
                    self.engine.format(
                        "MODIFY COLUMN {col_name:i} DROP NOT NULL",
                        {
                            "col_name": col_name,
                        },
                    )
                )
            elif not snow_c.not_null and bp_c.not_null:
                unsafe_alters.append(
                    self.engine.format(
                        "MODIFY COLUMN {col_name:i} SET NOT NULL",
                        {
                            "col_name": col_name,
                        },
                    )
                )

            # Default
            bp_c_default_value = self._normalize_bp_default(bp_c.default)

            if snow_c.default != bp_c_default_value:
                # DROP DEFAULT is supported
                if snow_c.default is not None and bp_c_default_value is None:
                    unsafe_alters.append(
                        self.engine.format(
                            "MODIFY COLUMN {col_name:i} DROP DEFAULT",
                            {
                                "col_name": col_name,
                            },
                        )
                    )

                # Switch to another sequence is supported
                elif (
                    isinstance(snow_c.default, str)
                    and snow_c.default.upper().endswith(".NEXTVAL")
                    and isinstance(bp_c_default_value, str)
                    and bp_c_default_value.upper().endswith(".NEXTVAL")
                ):
                    unsafe_alters.append(
                        self.engine.format(
                            "MODIFY COLUMN {col_name:i} SET DEFAULT {default:r}",
                            {
                                "col_name": col_name,
                                "default": bp_c_default_value,
                            },
                        )
                    )

                # All other DEFAULT changes are not supported
                else:
                    replace_reasons.append(f"Default for column {col_name} was changed")

            # Expression
            if snow_c.expression != bp_c.expression:
                replace_reasons.append(f"Expression for column {col_name} was changed")

            # Collate
            if snow_c.collate != bp_c.collate:
                replace_reasons.append(f"Collate for column {col_name} was changed")

            # Comments
            if snow_c.comment != bp_c.comment:
                if bp_c.comment:
                    safe_alters.append(
                        self.engine.format(
                            "MODIFY COLUMN {col_name:i} COMMENT {comment}",
                            {
                                "col_name": col_name,
                                "comment": bp_c.comment,
                            },
                        )
                    )
                else:
                    safe_alters.append(
                        self.engine.format(
                            "MODIFY COLUMN {col_name:i} UNSET COMMENT",
                            {
                                "col_name": col_name,
                            },
                        )
                    )

            # If type matches exactly, skip all other checks
            if snow_c.type == bp_c.type:
                continue

            # Only a few optimized MODIFY COLUMN ... TYPE actions are supported
            # https://docs.snowflake.com/en/sql-reference/sql/alter-table-column.html
            if snow_c.type.base_type == bp_c.type.base_type:
                # Increase or decrease precision of NUMBER, but not scale
                if (
                    snow_c.type.base_type == BaseDataType.NUMBER
                    and snow_c.type.val1 != bp_c.type.val2
                    and snow_c.type.val2 == bp_c.type.val2
                ):
                    unsafe_alters.append(
                        self.engine.format(
                            "MODIFY COLUMN {col_name:i} TYPE {col_type:r}",
                            {
                                "col_name": col_name,
                                "col_type": bp_c.type,
                            },
                        )
                    )

                    continue

                if snow_c.type.base_type == BaseDataType.VARCHAR and snow_c.type.val1 < bp_c.type.val1:
                    unsafe_alters.append(
                        self.engine.format(
                            "MODIFY COLUMN {col_name:i} TYPE {col_type:r}",
                            {
                                "col_name": col_name,
                                "col_type": bp_c.type,
                            },
                        )
                    )

                    continue

            # All other data type transformations require full table replace
            replace_reasons.append(f"Data type for column {col_name} was changed from {snow_c.type} to {bp_c.type}")

        # Remaining column names exactly match initial part of blueprint column names
        if remaining_col_names == list(islice(bp_cols.keys(), 0, len(remaining_col_names))):
            # Get remaining part of blueprint columns
            for col_name, bp_c in islice(bp_cols.items(), len(remaining_col_names), None):
                query = self.engine.query_builder()
                query.append(
                    "ADD COLUMN {col_name:i} {col_type:r}",
                    {
                        "col_name": col_name,
                        "col_type": bp_c.type,
                    },
                )

                if bp_c.collate:
                    query.append(
                        "COLLATE {collate}",
                        {
                            "collate": bp_c.collate,
                        },
                    )

                if bp_c.default is not None:
                    query.append(
                        "DEFAULT {default:r}",
                        {
                            "default": self._normalize_bp_default(bp_c.default),
                        },
                    )

                if bp_c.not_null:
                    query.append("NOT NULL")

                if bp_c.comment:
                    query.append(
                        "COMMENT {comment}",
                        {
                            "comment": bp_c.comment,
                        },
                    )

                safe_alters.append(query)
        else:
            # Reordering of columns is not supported
            replace_reasons.append("Order of columns was changed")

        # Changing TRANSIENT tables to permanent and back are not supported
        if bp.is_transient is True and row["is_transient"] is False:
            replace_reasons.append("Table type was changed to TRANSIENT")
        elif bp.is_transient is False and row["is_transient"] is True:
            replace_reasons.append("Table type was changed to PERMANENT")

        # Retention time
        if bp.retention_time is not None and bp.retention_time != row["retention_time"]:
            unsafe_alters.append(
                self.engine.format("SET DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}", {"retention_time": bp.retention_time})
            )

        # Clustering key
        if not self._compare_cluster_by(bp, row):
            if bp.cluster_by:
                unsafe_alters.append(
                    self.engine.format(
                        "CLUSTER BY ({cluster_by:r})",
                        {
                            "cluster_by": bp.cluster_by,
                        },
                    )
                )
            else:
                unsafe_alters.append(self.engine.format("DROP CLUSTERING KEY"))

        # Change tracking
        if bp.change_tracking != row["change_tracking"]:
            unsafe_alters.append(
                self.engine.format(
                    "SET CHANGE_TRACKING = {change_tracking:b}",
                    {
                        "change_tracking": bp.change_tracking,
                    },
                )
            )

        # Comment
        if bp.comment != row["comment"]:
            if bp.comment:
                safe_alters.append(
                    self.engine.format(
                        "SET COMMENT = {comment}",
                        {
                            "comment": bp.comment,
                        },
                    )
                )
            else:
                safe_alters.append(self.engine.format("UNSET COMMENT"))

        # Apply changes
        result = ResolveResult.NOCHANGE

        if replace_reasons:
            # fmt: off
            replace_query = "\n".join(f"-- {r}" for r in replace_reasons + replace_notices) + "\n" + str(self._build_create_table(bp, snow_cols))
            self.engine.execute_unsafe_ddl(replace_query, condition=self.engine.settings.execute_replace_table)
            # fmt: on

            result = ResolveResult.REPLACE

        elif safe_alters or unsafe_alters:
            for alter in safe_alters:
                self.engine.execute_safe_ddl(
                    "ALTER TABLE {full_name:i} {alter:r}",
                    {
                        "full_name": bp.full_name,
                        "alter": alter,
                    },
                )

            for alter in unsafe_alters:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} {alter:r}",
                    {
                        "full_name": bp.full_name,
                        "alter": alter,
                    },
                )

            result = ResolveResult.ALTER

            for col_name in dropping_col_names:
                self.engine.intention_cache.add_column_drop_intention(str(bp.full_name), col_name)

        # If table was re-created, apply or suggest search optimization using exactly the same condition value
        if result == ResolveResult.REPLACE:
            self._create_search_optimization(bp, condition=self.engine.settings.execute_replace_table)
        else:
            if self._compare_search_optimization(bp, row["search_optimization"]) and result == ResolveResult.NOCHANGE:
                result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP TABLE {database:i}.{schema:i}.{table_name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table_name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _get_existing_columns(self, bp: TableBlueprint):
        existing_columns = {}

        cur = self.engine.execute_meta(
            "DESC TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        for r in cur:
            m = collate_type_syntax_re.match(r["type"])

            if m:
                dtype = m.group(1)
                collate = m.group(2)
            else:
                dtype = r["type"]
                collate = None

            existing_columns[r["name"]] = TableColumn(
                name=Ident(r["name"]),
                type=DataType(dtype),
                not_null=bool(r["null?"] == "N"),
                default=r["default"] if r["default"] else None,
                expression=r["expression"] if r["expression"] else None,
                collate=collate,
                comment=r["comment"] if r["comment"] else None,
            )

        return existing_columns

    def _build_create_table(self, bp: TableBlueprint, snow_cols=None):
        query = self.engine.query_builder()
        query.append("CREATE")

        if snow_cols:
            query.append("OR REPLACE")

        if bp.is_transient:
            query.append("TRANSIENT")

        query.append(
            "TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append_nl("(")

        for idx, c in enumerate(bp.columns):
            query.append_nl(
                "    {comma:r}{col_name:i} {col_type:r}",
                {
                    "comma": "  " if idx == 0 else ", ",
                    "col_name": c.name,
                    "col_type": c.type,
                },
            )

            if c.collate:
                query.append(
                    "COLLATE {collate}",
                    {
                        "collate": c.collate,
                    },
                )

            if c.default is not None:
                query.append(
                    "DEFAULT {default:r}",
                    {
                        "default": self._normalize_bp_default(c.default),
                    },
                )

            if c.not_null:
                query.append("NOT NULL")

            if c.expression:
                query.append(
                    "AS ({expression:r})",
                    {
                        "expression": c.expression,
                    },
                )

            if c.comment:
                query.append(
                    "COMMENT {comment}",
                    {
                        "comment": c.comment,
                    },
                )

        query.append_nl(")")

        if bp.cluster_by:
            query.append_nl(
                "CLUSTER BY ({cluster_by:r})",
                {
                    "cluster_by": bp.cluster_by,
                },
            )

        if bp.change_tracking:
            query.append_nl("CHANGE_TRACKING = TRUE")

        if bp.retention_time is not None:
            query.append_nl("DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}", {"retention_time": bp.retention_time})

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        if snow_cols:
            query.append_nl("COPY GRANTS")
            query.append_nl("AS")
            query.append_nl("SELECT")

            # Skip virtual columns before enumeration
            for idx, c in enumerate(c for c in bp.columns if c.expression is None):
                col_name = str(c.name)

                if col_name in snow_cols:
                    # Column already exist

                    if c.type == snow_cols[col_name].type:
                        # Column has the same type
                        query.append_nl(
                            "    {comma:r}{col_name:i} AS {col_name:i}",
                            {
                                "comma": "  " if idx == 0 else ", ",
                                "col_name": c.name,
                            },
                        )
                    elif c.type.base_type == BaseDataType.VECTOR and snow_cols[col_name].type.base_type == BaseDataType.VECTOR:
                        # Change to another VECTOR type requires intermediate ARRAY
                        query.append_nl(
                            "    {comma:r}{col_name:i}::ARRAY::{col_type:r} AS {col_name:i}",
                            {
                                "comma": "  " if idx == 0 else ", ",
                                "col_name": c.name,
                                "col_type": c.type,
                            },
                        )
                    else:
                        # Column has a different type, add type cast
                        query.append_nl(
                            "    {comma:r}{col_name:i}::{col_type:r} AS {col_name:i}",
                            {
                                "comma": "  " if idx == 0 else ", ",
                                "col_name": c.name,
                                "col_type": c.type,
                            },
                        )
                else:
                    # Column does not exist, use default value with type cast
                    if c.default is not None:
                        query.append_nl(
                            "    {comma:r}{col_default:r}::{col_type:r} AS {col_name:i}",
                            {
                                "comma": "  " if idx == 0 else ", ",
                                "col_name": c.name,
                                "col_type": c.type,
                                "col_default": self._normalize_bp_default(c.default),
                            },
                        )
                    else:
                        query.append_nl(
                            "    {comma:r}NULL::{col_type:r} AS {col_name:i}",
                            {
                                "comma": "  " if idx == 0 else ", ",
                                "col_name": c.name,
                                "col_type": c.type,
                            },
                        )

            query.append_nl(
                "FROM {full_name:i}",
                {
                    "full_name": bp.full_name,
                },
            )

        return query

    def _compare_cluster_by(self, bp: TableBlueprint, row: dict):
        bp_cluster_by = ", ".join(bp.cluster_by) if bp.cluster_by else None
        snow_cluster_by = cluster_by_syntax_re.sub(r"\2", row["cluster_by"]) if row["cluster_by"] else None

        return bp_cluster_by == snow_cluster_by

    def _create_search_optimization(self, bp: TableBlueprint, condition=True):
        # Legacy search optimization on an entire table
        if isinstance(bp.search_optimization, bool):
            if bp.search_optimization:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} ADD SEARCH OPTIMIZATION",
                    {
                        "full_name": bp.full_name,
                    },
                    condition=condition,
                )

            return

        # Detailed search optimization on specific columns
        for bp_item in bp.search_optimization:
            self.engine.execute_unsafe_ddl(
                "ALTER TABLE {full_name:i} ADD SEARCH OPTIMIZATION ON {method:r}({target:i})",
                {
                    "full_name": bp.full_name,
                    "method": bp_item.method,
                    "target": bp_item.target,
                },
                condition=condition,
            )

    def _compare_search_optimization(self, bp: TableBlueprint, is_search_optimization_enabled=False, condition=True):
        # Legacy search optimization on an entire table
        if isinstance(bp.search_optimization, bool):
            if bp.search_optimization and not is_search_optimization_enabled:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} ADD SEARCH OPTIMIZATION",
                    {
                        "full_name": bp.full_name,
                    },
                    condition=condition,
                )

                return True

            elif not bp.search_optimization and is_search_optimization_enabled:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} DROP SEARCH OPTIMIZATION",
                    {
                        "full_name": bp.full_name,
                    },
                    condition=condition,
                )

                return True

            return False

        # Detailed search optimization on specific columns
        existing_search_optimization_items = []
        result = False

        cur = self.engine.execute_meta(
            "DESC SEARCH OPTIMIZATION ON {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        for r in cur:
            t_parts = r["target"].split(":", 2)

            existing_search_optimization_items.append(
                SearchOptimizationItem(
                    method=r["method"],
                    target=Ident(t_parts[0]),
                )
            )

        for bp_item in bp.search_optimization:
            if bp_item not in existing_search_optimization_items:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} ADD SEARCH OPTIMIZATION ON {method:r}({target:i})",
                    {
                        "full_name": bp.full_name,
                        "method": bp_item.method,
                        "target": bp_item.target,
                    },
                    condition=condition,
                )

                result = True

        for ex_item in existing_search_optimization_items:
            if ex_item not in bp.search_optimization:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} DROP SEARCH OPTIMIZATION ON {method:r}({target:i})",
                    {
                        "full_name": bp.full_name,
                        "method": ex_item.method,
                        "target": ex_item.target,
                    },
                    condition=condition,
                )

                result = True

        return result

    def _normalize_bp_default(self, bp_default):
        if isinstance(bp_default, SchemaObjectIdent):
            # This cannot be formatted properly with double-quotes ("), Snowflake strips it from DEFAULT value returned by DESC TABLE
            return f"{bp_default}.NEXTVAL"

        return bp_default



from logging import getLogger, NullHandler
from threading import get_ident as threading_get_ident

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from snowflake.connector import DictCursor, SnowflakeConnection, Error

from snowddl.cache import IntentionCache, SchemaCache
from snowddl.config import SnowDDLConfig
from snowddl.settings import SnowDDLSettings
from snowddl.formatter import SnowDDLFormatter
from snowddl.query_builder import SnowDDLQueryBuilder
from snowddl.context import SnowDDLContext
from snowddl.error import SnowDDLExecuteError


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class SnowDDLEngine:
    def __init__(self, connection: SnowflakeConnection, config: SnowDDLConfig, settings: SnowDDLSettings):
        self.connection = connection
        self.config = config
        self.settings = settings
        self.logger = logger

        self.formatter = SnowDDLFormatter()
        self.format = self.formatter.format_sql

        self.executor = ThreadPoolExecutor(max_workers=self.settings.max_workers, thread_name_prefix=self.__class__.__name__)

        self.executed_ddl = []
        self.suggested_ddl = []

        self._executed_ddl_buffer = defaultdict(list)
        self._suggested_ddl_buffer = defaultdict(list)

        self.context = SnowDDLContext(self)
        self.context.activate_role_with_prefix()

        self.intention_cache = IntentionCache(self)
        self.schema_cache = SchemaCache(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.executor.shutdown()

    def query_builder(self):
        return SnowDDLQueryBuilder(self.formatter)

    def describe_meta(self, sql, params=None):
        return self._describe(sql, params)

    def execute_meta(self, sql, params=None):
        return self._execute(sql, params, is_meta=True)

    def execute_clone(self, sql, params=None):
        return self._execute(sql, params)

    def execute_context_ddl(self, sql, params=None):
        return self._execute(sql, params)

    def execute_safe_ddl(self, sql, params=None, condition=True, file_stream=None):
        if self.settings.execute_safe_ddl and condition:
            self._execute(sql, params, False, file_stream)
        else:
            self._suggest(sql, params)

    def execute_unsafe_ddl(self, sql, params=None, condition=True, file_stream=None):
        if self.settings.execute_unsafe_ddl and condition:
            self._execute(sql, params, False, file_stream)
        else:
            self._suggest(sql, params)

    def flush_thread_buffers(self):
        for thread_sql in self._executed_ddl_buffer.values():
            for sql in thread_sql:
                self.executed_ddl.append(sql)

        for thread_sql in self._suggested_ddl_buffer.values():
            for sql in thread_sql:
                self.suggested_ddl.append(sql)

        self._executed_ddl_buffer = defaultdict(list)
        self._suggested_ddl_buffer = defaultdict(list)

    def _execute(self, sql, params, is_meta=False, file_stream=None):
        sql = self.format(sql, params)

        try:
            result = self.connection.cursor(DictCursor).execute(sql, file_stream=file_stream)
        except Error as e:
            raise SnowDDLExecuteError(e, sql)

        if not is_meta:
            self._executed_ddl_buffer[threading_get_ident()].append(sql)

        return result

    def _describe(self, sql, params):
        sql = self.format(sql, params)

        try:
            result = self.connection.cursor(DictCursor).describe(sql)
        except Error as e:
            raise SnowDDLExecuteError(e, sql)

        return result

    def _suggest(self, sql, params):
        sql = self.format(sql, params)
        self._suggested_ddl_buffer[threading_get_ident()].append(sql)
